"""Tools for describing a player character."""
__all__ = ('Character',)

import re
import os
import warnings
from . import exceptions
import importlib.util
import jinja2
import subprocess

from .stats import Ability, Skill, findattr, ArmorClass, Speed, Initiative
from .dice import read_dice_str
from . import (weapons, race, background, spells, armor, monsters,
               exceptions, classes, features, magic_items)
from .weapons import Weapon
from .armor import Armor, NoArmor, Shield, NoShield


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


__version__ = read('../VERSION').strip()


dice_re = re.compile('(\d+)d(\d+)')

__all__ = ('Barbarian', 'Bard', 'Cleric', 'Druid', 'Fighter', 'Monk',
           'Paladin', 'Ranger', 'Rogue', 'Sorcerer', 'Warlock', 'Wizard', )

multiclass_spellslots_by_level = {
    # char_lvl: (cantrips, 1st, 2nd, 3rd, ...)
    1:  (0, 2, 0, 0, 0, 0, 0, 0, 0, 0),
    2:  (0, 3, 0, 0, 0, 0, 0, 0, 0, 0),
    3:  (0, 4, 2, 0, 0, 0, 0, 0, 0, 0),
    4:  (0, 4, 3, 0, 0, 0, 0, 0, 0, 0),
    5:  (0, 4, 3, 2, 0, 0, 0, 0, 0, 0),
    6:  (0, 4, 3, 3, 0, 0, 0, 0, 0, 0),
    7:  (0, 4, 3, 3, 1, 0, 0, 0, 0, 0),
    8:  (0, 4, 3, 3, 2, 0, 0, 0, 0, 0),
    9:  (0, 4, 3, 3, 3, 1, 0, 0, 0, 0),
    10: (0, 4, 3, 3, 3, 2, 0, 0, 0, 0),
    11: (0, 4, 3, 3, 3, 2, 1, 0, 0, 0),
    12: (0, 4, 3, 3, 3, 2, 1, 0, 0, 0),
    13: (0, 4, 3, 3, 3, 2, 1, 1, 0, 0),
    14: (0, 4, 3, 3, 3, 2, 1, 1, 0, 0),
    15: (0, 4, 3, 3, 3, 2, 1, 1, 1, 0),
    16: (0, 4, 3, 3, 3, 2, 1, 1, 1, 0),
    17: (0, 4, 3, 3, 3, 2, 1, 1, 1, 1),
    18: (0, 4, 3, 3, 3, 3, 1, 1, 1, 1),
    19: (0, 4, 3, 3, 3, 3, 2, 1, 1, 1),
    20: (0, 4, 3, 3, 3, 3, 2, 2, 1, 1),
}


class Character():
    """A generic player character.
    
    """
    # General attirubtes
    name = ""
    player_name = ""
    alignment = "Neutral"
    dungeonsheets_version = __version__
    class_list = list()
    _race = None
    _background = None
    xp = 0
    # Hit points
    hp_max = 10
    # Base stats (ability scores)
    strength = Ability()
    dexterity = Ability()
    constitution = Ability()
    intelligence = Ability()
    wisdom = Ability()
    charisma = Ability()
    armor_class = ArmorClass()
    initiative = Initiative()
    speed = Speed()
    inspiration = 0
    _saving_throw_proficiencies = tuple()  # use to overwrite class proficiencies
    other_weapon_proficiencies = tuple()  # add to class/race proficiencies
    skill_proficiencies = list()
    skill_expertise = list()
    languages = ""
    # Skills
    acrobatics = Skill(ability='dexterity')
    animal_handling = Skill(ability='wisdom')
    arcana = Skill(ability='intelligence')
    athletics = Skill(ability='strength')
    deception = Skill(ability='charisma')
    history = Skill(ability='intelligence')
    insight = Skill(ability='wisdom')
    intimidation = Skill(ability='charisma')
    investigation = Skill(ability='intelligence')
    medicine = Skill(ability='wisdom')
    nature = Skill(ability='intelligence')
    perception = Skill(ability='wisdom')
    performance = Skill(ability='charisma')
    persuasion = Skill(ability='charisma')
    religion = Skill(ability='intelligence')
    sleight_of_hand = Skill(ability='dexterity')
    stealth = Skill(ability='dexterity')
    survival = Skill(ability='wisdom')
    # Characteristics
    attacks_and_spellcasting = ""
    personality_traits = "TODO: Describe how your character behaves, interacts with others"
    ideals = "TODO: Describe what values your character believes in."
    bonds = "TODO: Describe your character's commitments or ongoing quests."
    flaws = "TODO: Describe your character's interesting flaws."
    features_and_traits = "Describe any other features and abilities."
    # Inventory
    cp = 0
    sp = 0
    ep = 0
    gp = 0
    pp = 0
    equipment = ""
    weapons = list()
    magic_items = list()
    armor = None
    shield = None
    _proficiencies_text = list()
    # Magic
    spellcasting_ability = None
    _spells = list()
    _spells_prepared = list()
    # Features IN MAJOR DEVELOPMENT
    custom_features = list()
    feature_choices = list()
    
    def __init__(self, **attrs):
        """Takes a bunch of attrs and passes them to ``set_attrs``"""
        self.clear()
        # make sure class, race, background are set first
        my_classes = attrs.pop('classes', [])
        my_levels = attrs.pop('levels', [])
        my_subclasses = attrs.pop('subclasses', [])
        # backwards compatability
        if (len(my_classes) == 0):
            if ('class' in attrs):
                my_classes = [attrs.pop('class')]
                my_levels = [attrs.pop('level', 1)]
                my_subclasses = [attrs.pop('subclass', None)]
            else:  # if no classes or levels given, default to Lvl 1 Fighter
                my_classes = ['Fighter']
                my_levels = [1]
                my_subclasses = [None]
        # Generate the list of class objects
        self.add_classes(
            my_classes, my_levels, my_subclasses,
            feature_choices=attrs.get('feature_choices', []))
        # parse race and background
        self.race = attrs.pop('race', None)
        self.background = attrs.pop('background', None)
        # parse all other attributes
        self.set_attrs(**attrs)

    def clear(self):
        # reset class-definied items
        self.class_list = list()
        self.weapons = list()
        self.magic_items = list()
        self._saving_throw_proficiencies = tuple()
        self.other_weapon_proficiencies = tuple()
        self.skill_proficiencies = list()
        self.skill_expertise = list()
        self._proficiencies_text = list()
        self._spells = list()
        self._spells_prepared = list()
        self.custom_features = list()
        self.feature_choices = list()
                
    def __str__(self):
        return self.name
    
    def __repr__(self):
        return f"<{self.class_name}: {self.name}>"

    def add_class(self, cls: (classes.CharClass, type, str), level: (int, str),
                  subclass=None, feature_choices=[]):
        if isinstance(cls, str):
            cls = cls.strip().title().replace(' ', '')
            try:
                cls = getattr(classes, cls)
            except AttributeError:
                raise AttributeError(
                    'class was not recognized from classes.py: {:s}'.format(cls))
        if isinstance(level, str):
            level = int(level)
        self.class_list.append(cls(level, owner=self,
                                   subclass=subclass,
                                   feature_choices=feature_choices))
            
    def add_classes(self, classes_list=[], levels=[], subclasses=[],
                    feature_choices=[]):
        if isinstance(classes_list, str):
            classes_list = [classes_list]
        if isinstance(levels, int) or isinstance(levels, float) or isinstance(levels, str):
            levels = [levels]
        if len(levels) == 0:
            levels = [1]*len(classes_list)
        if isinstance(subclasses, str):
            subclasses = [subclasses]
        if len(subclasses) == 0:
            subclasses = [None]*len(classes_list)
        assert len(classes_list) == len(levels), (
            'the length of classes {:d} does not match length of '
            'levels {:d}'.format(len(classes), len(levels)))
        assert len(classes_list) == len(subclasses), (
            'the length of classes {:d} does not match length of '
            'subclasses {:d}'.format(len(classes_list), len(subclasses)))
        class_list = []
        for cls, lvl, sub in zip(classes_list, levels, subclasses):
            params = {}
            params['feature_choices'] = feature_choices
            self.add_class(cls=cls, level=lvl, subclass=sub,
                           **params)
    
    @property
    def race(self):
        return self._race

    @race.setter
    def race(self, newrace):
        if isinstance(newrace, race.Race):
            self._race = newrace
            self._race.owner = self
        elif isinstance(newrace, type) and issubclass(newrace, race.Race):
            self._race = newrace(owner=self)
        elif isinstance(newrace, str):
            try:
                self._race = findattr(race, newrace)(owner=self)
            except AttributeError:
                msg = (f'Race "{newrace}" not defined. '
                       f'Please add it to ``race.py``')
                self._race = race.Race(owner=self)
                warnings.warn(msg)
        elif newrace is None:
            self._race = race.Race(owner=self)
        
    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, bg):
        if isinstance(bg, background.Background):
            self._background = bg
            self._background.owner = self
        elif isinstance(bg, type) and issubclass(bg, background.Background):
            self._background = bg(owner=self)
        elif isinstance(bg, str):
            try:
                self._background = findattr(background, bg)(owner=self)
            except AttributeError:
                msg = (f'Background "{bg}" not defined. '
                       f'Please add it to ``background.py``')
                self._background = background.Background(owner=self)
                warnings.warn(msg)

    @property
    def class_name(self):
        if self.num_classes >= 1:
            return self.primary_class.name
        else:
            return ""

    @property
    def classes_and_levels(self):
        return ' / '.join([f'{c.name} {c.level}'
                           for c in self.class_list])

    @property
    def class_names(self):
        return [c.name for c in self.class_list]

    @property
    def levels(self):
        return [c.level for c in self.class_list]
    
    @property
    def subclasses(self):
        return list([c.subclass or '' for c in self.class_list])
    
    @property
    def level(self):
        return sum(c.level for c in self.class_list)

    @level.setter
    def level(self, new_level):
        self.primary_class.level = new_level
        if self.num_classes > 1:
            warnings.warn("Unable to tell which level to set. Updating "
                          "level of primary class {:s}".format(self.primary_class.name))
    
    @property
    def num_classes(self):
        return len(self.class_list)

    @property
    def has_class(self):
        return (self.num_classes > 0)

    @property
    def primary_class(self):
        # for now, assume first class given must be primary class
        if self.has_class:
            return self.class_list[0]
        else:
            return None

    @property
    def weapon_proficiencies(self):
        wp = set(self.other_weapon_proficiencies)
        if self.num_classes > 0:
            wp |= set(self.primary_class.weapon_proficiencies)
        if self.num_classes > 1:
            for c in self.class_list[1:]:
                wp |= set(c.multiclass_weapon_proficiencies)
        if self.race is not None:
            wp |= set(getattr(self.race, 'weapon_proficiencies', ()))
        if self.background is not None:
            wp |= set(getattr(self.background, 'weapon_proficiencies', ()))
        return tuple(wp)
    
    @weapon_proficiencies.setter
    def weapon_proficiencies(self, new_weapons):
        self.other_weapon_proficiencies = tuple(new_weapons)

    @property
    def other_weapon_proficiencies_text(self):
        return tuple(w.name for w in self.other_weapon_proficiencies)

    @property
    def features(self):
        fts = set(self.custom_features)
        if not self.has_class:
            return fts
        for c in self.class_list:
            fts |= set(c.features)
        if self.race is not None:
            fts |= set(getattr(self.race, 'features', ()))
            # some races have level-based features (Ex: Aasimar)
            if hasattr(self.race, 'features_by_level'):
                for lvl in range(1, self.level+1):
                    fts |= set(self.race.features_by_level[lvl])
        if self.background is not None:
            fts |= set(getattr(self.background, 'features', ()))
        return sorted(tuple(fts), key=(lambda x: x.name))

    @property
    def custom_features_text(self):
        return tuple([f.name for f in self.custom_features])

    def has_feature(self, feat):
        return any([isinstance(f, feat) for f in self.features])
    
    @property
    def saving_throw_proficiencies(self):
        if self.primary_class is None:
            return self._saving_throw_proficiencies
        else:
            return (self._saving_throw_proficiencies or
                    self.primary_class.saving_throw_proficiencies)

    @saving_throw_proficiencies.setter
    def saving_throw_proficiencies(self, vals):
        self._saving_throw_proficiencies = vals

    @property
    def spellcasting_classes(self):
        return [c for c in self.class_list if c.is_spellcaster]
    
    @property
    def is_spellcaster(self):
        return (len(self.spellcasting_classes) > 0)

    def spell_slots(self, spell_level):
        if len(self.spellcasting_classes) == 1:
            return self.spellcasting_classes[0].spell_slots(spell_level)
        else:
            if spell_level == 0:
                return sum([c.spell_slots(0)
                            for c in self.spellcasting_classes])
            else:
                # compute effective level from PHB pg 164
                eff_level = 0
                for c in self.spellcasting_classes:
                    if type(c) in [classes.Bard, classes.Cleric, classes.Druid,
                                   classes.Sorceror, classes.Wizard]:
                        eff_level += c.level
                    elif type(c) in [classes.Paladin, classes.Ranger]:
                        eff_level += c.level // 2
                    elif type(c) in [classes.Fighter, classes.Rogue]:
                        eff_level += c.level // 3
                if eff_level == 0:
                    return 0
                else:
                    return multiclass_spellslots_by_level[eff_level][spell_level]

    @property
    def spells(self):
        spells = set(self._spells) | set(self._spells_prepared)
        for f in self.features:
            spells |= set(f.spells_known) | set(f.spells_prepared)
        for c in self.spellcasting_classes:
            spells |= set(c.spells_known) | set(c.spells_prepared)
        if self.race is not None:
            spells |= set(self.race.spells_known) | set(self.race.spells_prepared)
        return sorted(tuple(spells), key=(lambda x: (x.name)))
            
    @property
    def spells_prepared(self):
        spells = set(self._spells_prepared)
        for f in self.features:
            spells |= set(f.spells_prepared)
        for c in self.spellcasting_classes:
            spells |= set(c.spells_prepared)
        if self.race is not None:
            spells |= set(self.race.spells_prepared)
        return sorted(tuple(spells), key=(lambda x: (x.name)))

    def set_attrs(self, **attrs):
        """Bulk setting of attributes. Useful for loading a character from a
        dictionary."""
        for attr, val in attrs.items():
            if attr == 'dungeonsheets_version':
                pass # Maybe we'll verify this later?
            elif attr == 'weapons':
                if isinstance(val, str):
                    val = [val]
                # Treat weapons specially
                for weap in val:
                    self.wield_weapon(weap)
            elif attr == 'magic_items':
                if isinstance(val, str):
                    val = [val]
                for mitem in val:
                    try:
                        self.magic_items.append(findattr(magic_items, mitem)(owner=self))
                    except (AttributeError):
                        msg = (f'Magic Item "{mitem}" not defined. '
                               f'Please add it to ``magic_items.py``')
                        warnings.warn(msg)
            elif attr == 'weapon_proficiencies':
                self.other_weapon_proficiencies = ()
                wps = set([findattr(weapons, w) for w in val])
                wps -= set(self.weapon_proficiencies)
                self.other_weapon_proficiencies = list(wps)
            elif attr == 'armor':
                self.wear_armor(val)
            elif attr == 'shield':
                self.wield_shield(val)
            elif attr == 'circle':
                if hasattr(self, 'Druid'):
                    self.Druid.circle = val
            elif attr == 'features':
                if isinstance(val, str):
                    val = [val]
                _features = []
                for f in val:
                    try:
                        _features.append(findattr(features, f))
                    except AttributeError:
                        msg = (f'Feature "{f}" not defined. '
                               f'Please add it to ``features.py``')
                        # create temporary feature
                        _features.append(features.create_feature(
                            name=f, source='Unknown',
                            __doc__="""Unknown Feature. Add to features.py"""))
                        warnings.warn(msg)
                self.custom_features += tuple(F(owner=self) for F in _features)
            elif (attr == 'spells') or (attr == 'spells_prepared'):
                # Create a list of actual spell objects
                _spells = []
                for spell_name in val:
                    try:
                        _spells.append(findattr(spells, spell_name))
                    except AttributeError:
                        msg = (f'Spell "{spell_name}" not defined. '
                               f'Please add it to ``spells.py``')
                        warnings.warn(msg)
                        # Create temporary spell
                        _spells.append(spells.create_spell(name=spell_name, level=9))
                        # raise AttributeError(msg)
                # Sort by name
                _spells.sort(key=lambda spell: spell.name)
                # Save list of spells to character atribute
                if attr == 'spells':
                    # Instantiate them all for the spells list
                    self._spells = tuple(S() for S in _spells)
                else:
                    # Instantiate them all for the spells list
                    self._spells_prepared = tuple(S() for S in _spells)
            else:
                if not hasattr(self, attr):
                    warnings.warn(f"Setting unknown character attribute {attr}",
                                  RuntimeWarning)
                # Lookup general attributes
                setattr(self, attr, val)

    def spell_save_dc(self, class_type):
        ability_mod = getattr(self, class_type.spellcasting_ability).modifier
        return (8 + self.proficiency_bonus + ability_mod)
    
    def spell_attack_bonus(self, class_type):
        ability_mod = getattr(self, class_type.spellcasting_ability).modifier
        return (self.proficiency_bonus + ability_mod)

    def is_proficient(self, weapon: Weapon):
        """Is the character proficient with this item?
        
        Considers class proficiencies and race proficiencies.
        
        Parameters
        ----------
        weapon
          The weapon to be tested for proficiency.

        Returns
        -------
        Boolean: is this character proficient with this weapon?
        
        """
        all_proficiencies = self.weapon_proficiencies
        is_proficient = any((isinstance(weapon, W) for W in all_proficiencies))
        return is_proficient
    
    @property
    def proficiencies_text(self):
        final_text = ""
        all_proficiencies = tuple(self._proficiencies_text)
        if self.has_class:
            all_proficiencies += tuple(self.primary_class._proficiencies_text)
        if self.num_classes > 1:
            for c in self.class_list[1:]:
                all_proficiencies += tuple(c._multiclass_proficiencies_text)
        if self.race is not None:
            all_proficiencies += tuple(self.race.proficiencies_text)
        if self.background is not None:
            all_proficiencies += tuple(self.background.proficiencies_text)
        # Create a single string out of all the proficiencies
        for txt in all_proficiencies:
            if not final_text:
                # Capitalize the first entry
                txt = txt.capitalize()
            else:
                # Put a comma first
                txt = ", " + txt
                # Add this item to the list text
            final_text += txt
        # Add a period at the end
        final_text += '.'
        return final_text

    @property
    def features_text(self):
        s = '\n\n--'.join([f.name + ("**" if f.needs_implementation else "")
                           for f in self.features])
        if s != '':
            s = '(See Features Page)\n\n--' + s
            s += '\n\n=================\n\n'
        return s
    
    @property
    def magic_items_text(self):
        s = ', '.join([f.name + ("**" if f.needs_implementation else "")
                        for f in sorted(self.magic_items, key=(lambda x: x.name))])
        if s:
            s += ', '
        return s
    
    def wear_armor(self, new_armor):
        """Accepts a string or Armor class and replaces the current armor.
        
        If a string is given, then a subclass of
        :py:class:`~dungeonsheets.armor.Armor` is retrived from the
        ``armor.py`` file. Otherwise, an subclass of
        :py:class:`~dungeonsheets.armor.Armor` can be provided
        directly.
        
        """
        if new_armor not in ('', 'None', None):
            if isinstance(new_armor, armor.Armor):
                new_armor = new_armor
            else:
                NewArmor = findattr(armor, new_armor)
                new_armor = NewArmor()
            self.armor = new_armor
    
    def wield_shield(self, shield):
        """Accepts a string or Shield class and replaces the current armor.
        
        If a string is given, then a subclass of
        :py:class:`~dungeonsheets.armor.Shield` is retrived from the
        ``armor.py`` file. Otherwise, an subclass of
        :py:class:`~dungeonsheets.armor.Shield` can be provided
        directly.
        
        """
        if shield not in ('', 'None', None):
            try:
                NewShield = findattr(armor, shield)
            except AttributeError:
                # Not a string, so just treat it as Armor
                NewShield = shield
            self.shield = NewShield()
    
    def wield_weapon(self, weapon):
        """Accepts a string and adds it to the list of wielded weapons.
        
        Parameters
        ----------
        weapon : str
          Case-insensitive string with a name of the weapon.
        
        """
        # Retrieve the weapon class from the weapons module
        if isinstance(weapon, weapons.Weapon):
            weapon_ = type(weapon)(wielder=self)
        elif isinstance(weapon, str):
            try:
                NewWeapon = findattr(weapons, weapon)
            except AttributeError:
                raise AttributeError(f'Weapon "{weapon}" is not defined')
            weapon_ = NewWeapon(wielder=self)
        elif issubclass(weapon, weapons.Weapon):
            weapon_ = weapon(wielder=self)
        else:
            raise AttributeError(f'Weapon "{weapon}" is not defined')
        # Save it to the array
        self.weapons.append(weapon_)
    
    @property
    def hit_dice(self):
        """What type and how many dice to use for re-gaining hit points.
        
        To change, set hit_dice_num and hit_dice_faces."""
        return ' + '.join([f'{c.level}d{c.hit_dice_faces}'
                           for c in self.class_list])

    @property
    def hit_dice_faces(self):
        # Not a valid function if multiclass
        if self.num_classes > 1:
            warnings.warn("hit_dice_faces is not valid for multiclass characters")
        return self.primary_class.hit_dice_faces
    
    @hit_dice_faces.setter
    def hit_dice_faces(self, faces):
        self.primary_class.hit_dice_faces = faces
    
    @property
    def proficiency_bonus(self):
        if self.level < 5:
            prof = 2
        elif 5 <= self.level < 9:
            prof = 3
        elif 9 <= self.level < 13:
            prof = 4
        elif 13 <= self.level < 17:
            prof = 5
        elif 17 <= self.level:
            prof = 6
        return prof
    
    def can_assume_shape(self, shape: monsters.Monster):
        return hasattr(self, 'Druid') and self.Druid.can_assume_shape(shape)

    @property
    def all_wild_shapes(self):
        if hasattr(self, 'Druid'):
            return self.Druid.all_wild_shapes
        else:
            return ()

    @property
    def wild_shapes(self):
        if hasattr(self, 'Druid'):
            return self.Druid.wild_shapes
        else:
            return ()

    @wild_shapes.setter
    def wild_shapes(self, new_shapes):
        if hasattr(self, 'Druid'):
            self.Druid.wild_shapes = new_shapes

    @classmethod
    def load(cls, character_file):
        # Create a character from the character definition
        char_props = read_character_file(character_file)
        classes = char_props.get('classes', [])
        # backwards compatability
        if (len(classes) == 0) and ('character_class' in char_props):
            char_props['classes'] = [char_props.pop('character_class').lower().capitalize()]
            char_props['levels'] = [str(char_props.pop('level'))]
        # Create the character with loaded properties
        char = Character(**char_props)
        return char

    def save(self, filename, template_file='character_template.txt'):
        # Create the template context
        context = dict(
            char=self,
        )
        # Render the template
        src_path = os.path.join(os.path.dirname(__file__), 'forms/')
        text = jinja2.Environment(
            loader=jinja2.FileSystemLoader(src_path or './')
        ).get_template(template_file).render(context)
        # Save the file
        with open(filename, mode='w') as f:
            f.write(text)

    def to_pdf(self, filename, **kwargs):
        from .make_sheets import make_sheet
        if filename.endswith('.pdf'):
            filename = filename.replace('pdf', 'py')
        make_sheet(filename, character=self,
                   flatten=kwargs.get('flatten', True))

        
def read_character_file(filename):
    """Create a character object from the given definition file.
    
    The definition file should be an importable python file, filled
    with variables describing the character.
    
    Parameters
    ----------
    filename : str
      The path to the file that will be imported.
    
    """
    # Parse the file name
    dir_, fname = os.path.split(os.path.abspath(filename))
    module_name, ext = os.path.splitext(fname)
    if ext != '.py':
        raise ValueError(f"Character definition {filename} is not a python file.")
    # Check if this file contains the version string
    version_re = re.compile('dungeonsheets_version\s*=\s*[\'"]([0-9.]+)[\'"]')
    with open(filename, mode='r') as f:
        version = None
        for line in f:
            match = version_re.match(line)
            if match:
                version = match.group(1)
                break
        if version is None:
            # Not a valid DND character file
            raise exceptions.CharacterFileFormatError(
                f"No ``dungeonsheets_version = `` entry in `{filename}`.")
    # Import the module to extract the information
    spec = importlib.util.spec_from_file_location('module', filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    # Prepare a list of properties for this character
    char_props = {}
    for prop_name in dir(module):
        if prop_name[0:2] != '__':
            char_props[prop_name] = getattr(module, prop_name)
    return char_props


# Add backwards compatability for tests
class Barbarian(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Barbarian']
        attrs['levels'] = [level]
        super().__init__(**attrs)
    

class Bard(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Bard']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Cleric(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Cleric']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Druid(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Druid']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Fighter(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Fighter']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Monk(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Monk']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Paladin(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Paladin']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Ranger(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Ranger']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Rogue(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Rogue']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Sorceror(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Sorceror']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Warlock(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Warlock']
        attrs['levels'] = [level]
        super().__init__(**attrs)


class Wizard(Character):
    def __init__(self, level=1, **attrs):
        attrs['classes'] = ['Wizard']
        attrs['levels'] = [level]
        super().__init__(**attrs)




from .spells import Spell


class PassWithoutTrace(Spell):
    """A veil of shadows and silence radiates from you, masking you and your companions
     from detection.
    For the duration, each creature you choose within 30 feet of 
    you (including you) has a +10 bonus to Dexterity (Stealth) checks and can’t be 
    tracked except by magical means. A creature that receives this bonus leaves 
    behind no tracks or other traces of its passage.
    """
    name = "Pass Without Trace"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """Ashes from a burned leaf of mistletoe and a sprig of spruce"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Druid', 'Ranger')


class Passwall(Spell):
    """A passage appears at a point of your choice that you can see on a wooden, 
    plaster, or stone surface (such as a wall, a ceiling, or a floor) within range, 
    and lasts for the duration. You choose the opening’s dimensions: up to 5 feet 
    wide, 8 feet tall, and 20 feet deep. The passage creates no instability in a 
    structure surrounding it.
    
    When the opening disappears, any creatures or objects
     still in the passage created by the spell are safely ejected to an unoccupied 
    space nearest to the surface on which you cast the spell.
    """
    name = "Passwall"
    level = 5
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A pinch of sesame seeds"""
    duration = "1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Wizard',)


class PhantasmalForce(Spell):
    """You craft an illusion that takes root in the mind of a creature that you can see
     within range.
    The target must make an Intelligence saving throw. On a failed 
    save, you create a phantasmal object, creature, or other visible phenomenon of 
    your choice that is no larger than a 10-foot cube and that is perceivable only 
    to the target for the duration. This spell has no effect on undead or 
    constructs.
    
    The phantasm includes sound, temperature, and other stimuli, also 
    evident only to the creature.
    
    The target can use its action to examine the 
    phantasm with an Intelligence (Investigation) check against your spell save DC. 
    If the check succeeds, the target realizes that the phantasm is an illusion, and
     the spell ends.
    
    While a target is affected by the spell, the target treats the
     phantasm as if it were real. The target rationalizes any illogical outcomes 
    from interacting with the phantasm. For example, a target attempting to walk 
    across a phantasmal bridge that spans a chasm falls once it steps onto the 
    bridge. If the target survives the fall, it still believes that the bridge 
    exists and comes up with some other explanation for its fallit was pushed, it 
    slipped, or a strong wind might have knocked it off.
    
    An affected target is so 
    convinced of the phantasm’s reality that it can even take damage from the 
    illusion. A phantasm created to appear as a creature can attack the target. 
    Similarly, a phantasm created to appear as fire, a pool of acid, or lava can 
    burn the target. Each round on your turn, the phantasm can deal 1d6 psychic 
    damage to the target if it is in the phantasm’s area or within 5 feet of the 
    phantasm, provided that the illusion is of a creature or hazard that could 
    logically deal damage, such as by attacking. The target perceives the damage as 
    a type appropriate to the illusion.
    """
    name = "Phantasmal Force"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A bit of fleece"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class PhantasmalKiller(Spell):
    """You tap into the nightmares of a creature you can see within range and create an
     illusory manifestation of its deepest fears, visible only to that creature.
    The
     target must make a Wisdom saving throw. On a failed save, the target becomes 
    frightened for the duration. At the end of each of the target’s turns before the
     spell ends, the target must succeed on a Wisdom saving throw or take 4d10 
    psychic damage. On a successful save, the spell ends.
    
    At Higher Levels: When 
    you cast this spell using a spell slot of 5th level or higher, the damage 
    increases by 1d1O for each slot level above 4th.
    """
    name = "Phantasmal Killer"
    level = 4
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Wizard',)


class PhantomSteed(Spell):
    """A Large quasi-real, horselike creature appears on the ground in an unoccupied 
    space of your choice within range. You decide the creature’s appearance, but it 
    is equipped with a saddle, bit, and bridle. Any of the equipment created by the 
    spell vanishes in a puff of smoke if it is carried more than 10 feet away from 
    the steed.
    
    For the duration, you or a creature you choose can ride the steed. 
    The creature uses the statistics for a riding horse, except it has a speed of 
    100 feet and can travel 10 miles in an hour, or 13 miles at a fast pace. When 
    the spell ends, the steed gradually fades, giving the rider 1 minute to 
    dismount. The spell ends if you use an action to dismiss it or if the steed 
    takes any damage.
    """
    name = "Phantom Steed"
    level = 3
    casting_time = "1 minute"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "1 hour"
    ritual = True
    magic_school = "Illusion"
    classes = ('Wizard',)


class PlanarAlly(Spell):
    """You beseech an otherworldly entity for aid.
    The being must be known to you: a 
    god, a primordial, a demon prince, or some other being of cosmic power. That 
    entity sends a celestial, an elemental, or a fiend loyal to it to aid you, 
    making the creature appear in an unoccupied space within range. If you know a 
    specific creature’s name, you can speak that name when you cast this spell to 
    request that creature, though you might get a different creature anyway (DM’s 
    choice).
    
    When the creature appears, it is under no compulsion to behave in any 
    particular way. You can ask the creature to perform a service in exchange for 
    payment, but it isn’t obliged to do so. The requested task could range from 
    simple (fly us across the chasm, or help us fight a battle) to complex (spy on 
    our enemies, or protect us during our foray into the dungeon). You must be able 
    to communicate with the creature to bargain for its services.
    
    Payment can take 
    a variety of forms. A celestial might require a sizable donation of gold or 
    magic items to an allied temple, while a fiend might demand a living sacrifice 
    or a gift of treasure. Some creatures might exchange their service for a quest 
    undertaken by you.
    
    As a rule of thumb, a task that can be measured in minutes 
    requires a payment worth 100 gp per minute. A task measured in hours requires 
    1,000 gp per hour. And a task m easured in days (up to 10 days) requires 10,000 
    gp per day. The DM can adjust these payments based on the circumstances under 
    which you cast the spell. If the task is aligned with the creature’s ethos, the 
    payment might be halved or even waived. Nonhazardous tasks typically require 
    only half the suggested payment, while especially dangerous tasks might require 
    a greater gift. Creatures rarely accept tasks that seem suicidal.
    
    After the 
    creature completes the task, or when the agreed-upon duration of service 
    expires, the creature returns to its home plane after reporting back to you, if 
    appropriate to the task and if possible. If you are unable to agree on a price 
    for the creature’s service, the creature immediately returns to its home plane.
    
    
    A creature enlisted to join your group counts as a member of it, receiving a 
    full share of experience points awarded.
    """
    name = "Planar Ally"
    level = 6
    casting_time = "10 minutes"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Cleric',)


class PlanarBinding(Spell):
    """With this spell, you attempt to bind a celestial, an elemental, a fey, or a 
    fiend to your service.
    The creature must be within range for the entire casting 
    of the spell. (Typically, the creature is first summoned into the center of an 
    inverted magic circle in order to keep it trapped while this spell is cast.) At 
    the completion of the casting, the target must make a Charisma saving throw. On 
    a failed save, it is bound to serve you for the duration. If the creature w as 
    summoned or created by another spell, that spell’s duration is extended to match
     the duration of this spell.
    
    A bound creature must follow your instructions to 
    the best of its ability. You might command the creature to accompany you on an 
    adventure, to guard a location, or to deliver a message. The creature obeys the 
    letter of your instructions, but if the creature is hostile to you, it strives 
    to twist your words to achieve its own objectives. If the creature carries out 
    your instructions completely before the spell ends, it travels to you to report 
    this fact if you are on the same plane of existence. If you are on a different 
    plane of existence, it returns to the place where you bound it and remains there
     until the spell ends.
    
    At Higher Levels: When you cast this spell using a spell
     slot of a higher level, the duration increases to:
    10 days with a 6th-level 
    slot,
    30 days with a 7th-level slot,
    180 days with an 8th-level slot, 
    1 year
     and 1 day with a 9th-level spell slot.
    """
    name = "Planar Binding"
    level = 5
    casting_time = "1 hour"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A jewel worth at least 1,000 gp, which the spell consumes"""
    duration = "24 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Bard', 'Cleric', 'Druid', 'Wizard')


class PlaneShift(Spell):
    """You and up to eight willing creatures who link hands in a circle are transported
     to a different plane of existence. You can specify a target destination in 
    general terms, such as the City of Brass on the Elemental Plane of Fire or the 
    palace of Dispater on the second level of the Nine Hells, and you appear in or 
    near that destination. If you are trying to reac the City of Brass, for example,
     you might arrive in its Street of Steel, before its Gate of Ashes, or looking 
    at the city from across the Sea of Fire, at the DM’s discretion.
    
    Alternatively,
     if you know the sigil sequence of a teleportation circle on another plane of 
    existence, this spell can take you to that circle. If the teleportation circle 
    is too small to hold all the creatures you transported, they appear in the 
    closest unoccupied spaces next to the circle.
    
    You can use this spell to banish 
    an unwilling creature to another plane. Choose a creature within your reach and 
    make a melee spell attack against it. On a hit, the creature must make a 
    Charisma saving throw. If the creature fails the save, it is transported to a 
    random location on the plane of existence you specify. A creature so transported
     must find its own way back to your current plane of existence.
    """
    name = "Plane Shift"
    level = 7
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A forked, metal rod worth at least 250 gp, attuned to a particular plane of existence"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Cleric', 'Druid', 'Sorcerer', 'Warlock', 'Wizard')


class PlantGrowth(Spell):
    """This spell channels vitality into plants within a specific area. There are two 
    possible uses for the spell, granting either immediate or long-term benefits.
    
    
    If you cast this spell using 1 action, choose a point within range. All normal 
    plants in a 100-foot radius centered on that point become thick and overgrown. A
     creature moving through the area must spend 4 feet of movement for every 1 foot
     it moves.
    
    You can exclude one or more areas of any size within the spell’s 
    area from being affected.
    
    If you cast this spell over 8 hours, you enrich the 
    land. All plants in a half-mile radius centered on a point within range become 
    enriched for 1 year. The plants yield twice the normal amount of food when 
    harvested.
    """
    name = "Plant Growth"
    level = 3
    casting_time = "Special"
    casting_range = "150 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Druid', 'Ranger')


class PoisonSpray(Spell):
    """You extend your hand toward a creature you can see within range and project a 
    puff of noxious gas from your palm. The creature must succeed on a Constitution 
    saving throw or take 1d12 poison damage.
    
    At Higher Levels: This spell’s damage 
    increases by 1d12 when you reach 5th level (2d12), 11th level (3d12), 17th level
     (4d12).
    """
    name = "Poison Spray"
    level = 0
    casting_time = "1 action"
    casting_range = "10 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Druid', 'Sorcerer', 'Warlock', 'Wizard')


class Polymorph(Spell):
    """This spell transforms a creature that you can see within range into a new form. 
    An unwilling creature must make a Wisdom saving throw to avoid the effect. A 
    shapechanger automatically succeeds on this saving throw.
    
    The transformation 
    lasts for the duration, or until the target drops to 0 hit points or dies. The 
    new form can be any beast whose challenge rating is equal to or less than the 
    target’s (or the target’s level, if it doesn’t have a challenge rating). The 
    target’s game statistics, including mental ability scores, are replaced by the 
    statistics of the chosen beast. It retains its alignment and personality.
    
    The 
    target assumes the hit points of its new form. When it reverts to its normal 
    form, the creature returns to the number of hit points it had before it 
    transformed. If it reverts as a result of dropping to 0 hit points, any excess 
    damage carries over to its normal form. As long as the excess damage doesn’t 
    reduce the creature’s normal form to 0 hit points, it isn’t knocked unconscious.
    
    
    The creature is limited in the actions it can perform by the nature of its new
     form, and it can’t speak, cast spells, or take any other action that requires 
    hands or speech.
    
    The target’s gear melds into the new form. The creature can’t 
    activate, use, wield, or otherwise benefit from any of its equipment. This spell
     can’t affect a target that has 0 hit points.
    """
    name = "Polymorph"
    level = 4
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A caterpillar cocoon"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Druid', 'Sorcerer', 'Wizard')


class PowerWordHeal(Spell):
    """A wave of healing energy washes over the creature you touch. The target regains 
    all its hit points. If the creature is charmed, frightened, paralyzed, or 
    stunned, the condition ends. If the creature is prone, it can use its reaction 
    to stand up. This spell has no effect on undead or constructs.
    """
    name = "Power Word Heal"
    level = 9
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Bard',)


class PowerWordKill(Spell):
    """You utter a word of power that can compel one creature you can see within range 
    to die instantly. If the creature you chose has 100 hit points or fewer, it 
    dies. Otherwise, the spell has no effect.
    """
    name = "Power Word Kill"
    level = 9
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class PowerWordPain(Spell):
    """You speak a word of power that causes waves of intense pain to assail one 
    creature you can see within range. If the target has 100 hit points or fewer, it
     is subject to crippling pain. Otherwise, the spell has no effect on it. A 
    target is also unaffected if it is immune to being charmed.
    While the target is 
    affected by crippling pain, any speed it has can be no higher than 10 feet. The 
    target also has disadvantage on attack rolls, ability checks, and saving throws,
     other than Constitution saving throws. Finally, if the target tries to cast a 
    spell, it must first succeed on a Constitution saving throw, or the casting 
    fails and the spell is wasted.
    A target suffering this pain can make a 
    Constitution saving throw at the end of each of its turns. On a successful save,
     the pain ends.
    """
    name = "Power Word Pain"
    level = 7
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class PowerWordStun(Spell):
    """You speak a word of power that can overwhelm the mind of one creature you can 
    see within range, leaving it dumbfounded. If the target has 150 hit points or 
    fewer, it is stunned. Otherwise, the spell has no effect. The stunned target 
    must make a Constitution saving throw at the end of each of its turns. On a 
    successful save, this stunning effect ends.
    """
    name = "Power Word Stun"
    level = 8
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class PrayerOfHealing(Spell):
    """Up to six creatures of your choice that you can see within range each regain hit
     points equal to 2d8 + your spellcasting ability modifier. This spell has no 
    effect on undead or constructs.
    
    At Higher Levels: When you cast this spell 
    using a spell slot of 3rd level or higher, the healing increases by 1d8 for each
     slot level above 2nd.
    """
    name = "Prayer Of Healing"
    level = 2
    casting_time = "10 minutes"
    casting_range = "30 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric',)


class Prestidigitation(Spell):
    """This spell is a minor magical trick that novice spellcasters use for practice. 
    You create one of the following magical effects within range:
     -You create an 
    instantaneous, harmless sensory effect, such as a shower of sparks, a puff of 
    wind, faint musical notes, or an odd odor. 
    -You instantaneously light or snuff 
    out a candle, a torch, or a small campfire. 
    -You instantaneously clean or soil 
    an object no larger than 1 cubic foot. 
    -You chill, warm, or flavor up to 1 
    cubic foot of nonliving material for 1 hour. 
    -You make a color, a small mark, 
    or a symbol appear on an object or a surface for 1 hour. 
    -You create a 
    nonmagical trinket or an illusory image that can fit in your hand and that lasts
     until the end of your next turn. 
    If you cast this spell multiple times, you 
    can have up to three of its non-instantaneous effects active at a time, and you 
    can dismiss such an effect as an action.
    """
    name = "Prestidigitation"
    level = 0
    casting_time = "1 action"
    casting_range = "10 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class PrimalSavagery(Spell):
    """You channel primal magic to cause your teeth or fingernails to sharpen, ready to
     deliver a corrosive attack. Make a melee spell attack against one creature 
    within 5 feet of you. On a hit, the target takes 1d10 acid damage. After you 
    make the attack, your teeth or fingernails return to normal. The spell’s damage 
    increases by 1d10 when you reach 5th level (2d10), 11th level (3d10), and 17th 
    level (4d10).
    """
    name = "Primal Savagery"
    level = 0
    casting_time = "1 action"
    casting_range = "Self"
    components = ('S',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid',)


class PrimordialWard(Spell):
    """You have resistance to acid, cold, fire, lightning, and thunder damage for the 
    spell’s duration.
    When you take damage of one of those types, you can use your 
    reaction to gain immunity to that type
    of damage, including against the 
    triggering damage. If you do so, the resistances end, and you have the immunity 
    until the end of your next turn, at which time the spell ends.
    """
    name = "Primordial Ward"
    level = 6
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Druid',)


class PrismaticSpray(Spell):
    """Eight multicolored rays of light flash from your hand. Each ray is a different 
    color and has a different power and purpose. Each creature in a 60-foot cone 
    must make a Dexterity saving throw. For each target, roll a d8 to determine 
    which color ray affects it.
    
    1. Red. The target takes 10d6 fire damage on a 
    failed save, or half as much damage on a successful one.
    
    2. Orange. The target 
    takes 10d6 acid damage on a failed save, or half as much damage on a successful 
    one.
    
    3. Yellow. The target takes 10d6 lightning damage on a failed save, or 
    half as much damage on a successful one.
    
    4. Green. The target takes 10d6 poison
     damage on a failed save, or half as much damage on a successful one.
    
    5. Blue. 
    The target takes 10d6 cold damage on a failed save, or half as much damage on a 
    successful one.
    
    6. Indigo. On a failed save, the target is restrained. It must 
    then make a Constitution saving throw at the end of each of its turns. If it 
    successfully saves three times, the spell ends. If it fails its save three 
    times, it permanently turns to stone and is subjected to the petrified 
    condition. The successes and failures don’t need to be consecutive; keep track 
    of both until the target collects three of a kind.
    
    7. Violet. On a failed save,
     the target is blinded. It must then make a Wisdom saving throw at the start of 
    your next turn. A successful save ends the blindness. If it fails that save, the
     creature is transported to another plane of existence of the DM’s choosing and 
    is no longer blinded. (Typically, a creature that is on a plane that isn’t its 
    home plane is banished home, while other creatures are usually cast into the 
    Astral or Ethereal planes.)
    
    8. Special. The target is struck by two rays. Roll 
    twice more, rerolling any 8.
    """
    name = "Prismatic Spray"
    level = 7
    casting_time = "1 action"
    casting_range = "Self (60-foot cone)"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class PrismaticWall(Spell):
    """A shimmering, multicolored plane of light forms a vertical opaque wallup to 90 
    feet long, 30 feet high, and 1 inch thickcentered on a point you can seewithin 
    range. Alternatively, you can shape the wall into a sphere up to 30 feet in 
    diameter centered on a point you choose within range. The wall remains in place 
    for the duration. If you position the wall so that it passes through a space 
    occupied by a creature, the spell fails, and your action and the spell slot are 
    wasted.
    
    The wall sheds bright light out to a range of 100 feet and dim light 
    for an additional 100 feet. You and creatures you designate at the time you cast
     the spell can pass through and remain near the wall without harm. If another 
    creature that can see the wall moves to within 20 feet of it or starts its turn 
    there, the creature must succeed on a Constitution saving throw or become 
    blinded for 1 minute.
    
    The wall consists of seven layers, each with a different 
    color. When a creature attempts to reach into or pass through the wall, it does 
    so one layer at a time through all the wall’s layers. As it passes or reaches 
    through each layer, the creature must make a Dexterity saving throw or be 
    affected by that layer’s properties as described below.
    
    The wall can be 
    destroyed, also one layer at a time, in order from red to violet, by means 
    specific to each layer. Once a layer is destroyed, it remains so for the 
    duration of the spell. A rod of cancellation destroys a prismatic wall, but an 
    antimagic field has no effect on it.
    
    1. Red. The creature takes 10d6 fire 
    damage on a failed save, or half as much damage on a successful one. While this 
    layer is in place, nonmagical ranged attacks can’t pass through the wall. The 
    layer can be destroyed by dealing at least 25 cold damage to it.
    
    2. Orange. The
     creature takes 10d6 acid damage on a failed save, or half as much damage on a 
    successful one. While this layer is in place, magical ranged attacks can’t pass 
    through the wall. The layer is destroyed by a strong wind.
    
    3. Yellow. The 
    creature takes 10d6 lightning damage on a failed save, or half as much damage on
     a successful one. This layer can be destroyed by dealing at least 60 force 
    damage to it.
    
    4. Green. The creature takes 10d6 poison damage on a failed save,
     or half as much damage on a successful one. A passwall spell, or another spell 
    of equal or greater level that can open a portal on a solid surface, destroys 
    this layer.
    
    5. Blue. The creature takes 10d6 cold damage on a failed save, or 
    half as much damage on a successful one. This layer can be destroyed by dealing 
    at least 25 fire damage to it.
    
    6. Indigo. On a failed save, the creature is 
    restrained. It must then make a Constitution saving throw at the end of each of 
    its turns. If it successfully saves three times, the spell ends. If it fails its
     save three times, it permanently turns to stone and is subjected to the 
    petrified condition. The successes and failures don’t need to be consecutive; 
    keep track of both until the creature collects three of a kind. While this layer
     is in place, spells can’t be cast through the wall. The layer is destroyed by 
    bright light shed by a daylight spell or a similar spell of equal or higher 
    level.
    
    7. Violet. On a failed save, the creature is blinded. It must then make 
    a Wisdom saving throw at the start of your next turn. A successful save ends the
     blindness. If it fails that save, the creature is transported to another plane 
    of the DM’s choosing and is no longer blinded. (Typically, a creature that is on
     a plane that isn’t its home plane is banished home, while other creatures are 
    usually cast into the Astral or Ethereal planes.) This layer is destroyed by a 
    dispel magic spell or similar spell of equal or higher level that can end spells
     and magical effects.
    """
    name = "Prismatic Wall"
    level = 9
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Wizard',)


class ProduceFlame(Spell):
    """A flickering flame appears in your hand.
    The flame remains there for the 
    duration and harms neither you nor your equipment. The flame sheds bright light 
    in a 10-foot radius and dim light for an additional 10 feet. The spell ends if 
    you dismiss it as an action or if you cast it again.
    
    You can also attack with 
    the flame, although doing so ends the spell. When you cast this spell, or as an 
    action on a later turn, you can hurl the flame at a creature within 30 feet of 
    you. Make a ranged spell attack. On a hit, the target takes 1d8 fire damage.
    
    At
     Higher Levels: This spell’s damage increases by 1d8 when you reach 5th level 
    (2d8), 11th level (3d8), and 17th level (4d8).
    """
    name = "Produce Flame"
    level = 0
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Druid',)


class ProgrammedIllusion(Spell):
    """You create an illusion of an object, a creature, or some other visible 
    phenomenon within range that activates when a specific condition occurs. The 
    illusion is imperceptible until then. It must be no larger than a 30-foot cube, 
    and you decide when you cast the spell how the illusion behaves and what sounds 
    it makes. This scripted performance can last up to 5 minutes.
    
    When the 
    condition you specify occurs, the illusion springs into existence and performs 
    in the manner you described. Once the illusion finishes performing, it 
    disappears and remains dormant for 10 minutes. After this time, the illusion can
     be activated again.
    
    The triggering condition can be as general or as detailed 
    as you like, though it must be based on visual or audible conditions that occur 
    within 30 feet of the area. For example, you could create an illusion of 
    yourself to appear and warn off others who attempt to open a trapped door, or 
    you could set the illusion to trigger only when a creature says the correct word
     or phrase.
    
    Physical interaction with the image reveals it to be an illusion, 
    because things can pass through it. A creature that uses its action to examine 
    the image can determine that it is an illusion with a successful Intelligence 
    (Investigation) check against your spell save DC. If a creature discerns the 
    illusion for what it is, the creature can see through the image, and any noise 
    it makes sounds hollow to the creature.
    """
    name = "Programmed Illusion"
    level = 6
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A bit of fleece and jade dust worth at least 25 gp"""
    duration = "Until dispelled"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Wizard')


class ProjectImage(Spell):
    """You create an illusory copy of yourself that lasts for the duration.
    The copy 
    can appear at any location within range that you have seen before, regardless of
     intervening obstacles. The illusion looks and sounds like you but is 
    intangible. If the illusion takes any damage, it disappears, and the spell ends.
    
    
    You can use your action to move this illusion up to twice your speed, and make
     it gesture, speak, and behave in whatever way you choose. It mimics your 
    mannerisms perfectly.
    
    You can see through its eyes and hear through its ears as
     if you were in its space. On your turn as a bonus action, you can switch from 
    using its senses to using your own, or back again. While you are using its 
    senses, you are blinded and deafened in regard to your own surroundings.
    
    
    Physical interaction with the image reveals it to be an illusion, because things
     can pass through it. A creature that uses its action to examine the image can 
    determine that it is an illusion with a successful Intelligence (Investigation) 
    check against your spell save DC. If a creature discerns the illusion for what 
    it is, the creature can see through the image, and any noise it makes sounds 
    hollow to the creature.
    """
    name = "Project Image"
    level = 7
    casting_time = "1 action"
    casting_range = "500 miles"
    components = ('V', 'S', 'M')
    materials = """A small replica of you made from materials worth at least 5 gp"""
    duration = "Concentration, up to 1 day"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Wizard')


class ProtectionFromEnergy(Spell):
    """For the duration, the willing creature you touch has resistance to one damage 
    type of your choice: acid, cold, fire, lightning, or thunder.
    """
    name = "Protection From Energy"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric', 'Druid', 'Ranger', 'Sorcerer', 'Wizard')


class ProtectionFromEvilAndGood(Spell):
    """Until the spell ends, one willing creature you touch is protected against 
    certain types of creatures: aberrations, celestials, elementals, fey, fiends, 
    and undead.
    
    The protection grants several benefits. Creatures of those types 
    have disadvantage on attack rolls against the target. The target also can’t be 
    charmed, frightened, or possessed by them. If the target is already charmed, 
    frightened, or possessed by such a creature, the target has advantage on any new
     saving throw against the relevant effect.
    """
    name = "Protection From Evil And Good"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Holy water or powdered silver and iron, which the spell consumes"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric', 'Paladin', 'Warlock', 'Wizard')


class ProtectionFromPoison(Spell):
    """You touch a creature. If it is poisoned, you neutralize the poison. If more than
     one poison afflicts the target, you neutralize one poison that you know is 
    present, or you neutralize one at random.
    
    For the duration, the target has 
    advantage on saving throws against being poisoned, and it has resistance to 
    poison damage.
    """
    name = "Protection From Poison"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric', 'Druid', 'Paladin', 'Ranger')


class PsychicScream(Spell):
    """You unleash the power of your mind to blast the intellect of up to ten creatures
     of your choice that you can see within range. Creatures that have an 
    Intelligence score of 2 or lower are unaffected.
    Each target must make an 
    Intelligence saving throw. On a failed save, a target takes 14d6 psychic damage 
    and is stunned. On a successful save, a target takes half as much damage and 
    isn’t stunned. If a target is killed by this damage, its head explodes, assuming
     it has one.
    A stunned target can make an Intelligence saving throw at the end 
    of each of its turns. On a successful save, the stunning effect ends.
    """
    name = "Psychic Scream"
    level = 9
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('S',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class PurifyFoodAndDrink(Spell):
    """All nonmagical food and drink within a 5-foot-radius sphere centered on a point 
    of your choice within range is purified and rendered free of poison and disease.
    """
    name = "Purify Food And Drink"
    level = 1
    casting_time = "1 action"
    casting_range = "10 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = True
    magic_school = "Transmutation"
    classes = ('Cleric', 'Druid', 'Paladin')


class Pyrotechnics(Spell):
    """Choose an area of nonmagical flame that you can see and that fits within a 
    5-foot cube within range. You can extinguish the fire in that area, and you 
    create either fireworks or smoke when you do so.
    Fireworks. The target explodes 
    with a dazzling display of colors. Each creature within 10 feet of the target 
    must succeed on a Constitution saving throw or become blinded until the end of 
    your next turn.
    Smoke. Thick black smoke spreads out from the target in a 
    20-foot radius, moving around corners. The area of the smoke is heavily 
    obscured. The smoke persists for 1 minute or until a strong wind disperses it.
    """
    name = "Pyrotechnics"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Sorcerer', 'Wizard')





from .spells import Spell


class GaseousForm(Spell):
    """You transform a willing creature you touch, along with everything it’s wearing 
    and carrying, into a misty cloud for the duration. The spell ends if the 
    creature drops to 0 hit points. An incorporeal creature isn’t affected.
    
    While 
    in this form, the target’s only method of movement is a flying speed of 10 feet.
     The target can enter and occupy the space of another creature. The target has 
    resistance to nonmagical damage, and it has advantage on Strength, Dexterity, 
    and Constitution saving throws. The target can pass through small holes, narrow 
    openings, and even mere cracks, though it treats liquids as though they were 
    solid surfaces. The target can’t fall and remains hovering in the air even when 
    stunned or otherwise incapacitated.
    
    While in the form of a misty cloud, the 
    target can’t talk or manipulate objects, and any objects it was carrying or 
    holding can’t be dropped, used, or otherwise interacted with. The target can’t 
    attack or cast spells.
    """
    name = "Gaseous Form"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A bit of gauze and a wisp of smoke"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class Gate(Spell):
    """You conjure a portal linking an unoccupied space you can see within range to a 
    precise location on a different plane of existence. The portal is a circular 
    opening, which you can make 5 to 20 feet in diameter. You can orient the portal 
    in any direction you choose. The portal lasts for the duration.
    
    The portal has 
    a front and a back on each plane where it appears. Travel through the portal is 
    possible only by moving through its front. Anything that does so is instantly 
    transported to the other plane, appearing in the unoccupied space nearest to the
     portal.
    
    Deities and other planar rulers can prevent portals created by this 
    spell from opening in their presence or anywhere within their domains.
    
    When you
     cast this spell, you can speak the name of a specific creature (a pseudonym, 
    title, or nickname doesn’t work). If that creature is on a plane other than the 
    one you are on, the portal opens in the named creature’s immediate vicinity and 
    draws the creature through it to the nearest unoccupied space on your side of 
    the portal. You gain no special power over the creature, and it is free to act 
    as the Dm deems appropriate. It might leave, attack you, or help you.
    """
    name = "Gate"
    level = 9
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A diamond worth at least 5,000 gp"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Cleric', 'Sorcerer', 'Wizard')


class Geas(Spell):
    """You place a magical command on a creature that you can see within range, forcing
     it to carry out some service or refrain from some action or course of actiity 
    as you decide.
    If the creature can understand you, it must succeed on a Wisdom 
    saving throw or become charmed by you for the duration. While the creature is 
    charmed by you, it takes 5d10 psychic damage each time it acts in a manner 
    directly counter to your instructions, but no more than once each day. A 
    creature that can’t understand you is unaffected by the spell.
    
    You can issue 
    any command you choose, short of an activity that would result in certain death.
     Should you issue a suicidal command, the spell ends. You can end the spell 
    early by using an action to dismiss it. A remove curse, greater restoration, or 
    wish spell also ends it.
    
    At Higher Levels: When you cast this spell usinga 
    spell slot of 7th or 8th level, the duration is 1 year.
    When you cast this 
    spell using a spell slot of 9th level, the spell lasts until it is ended by one 
    of the spells mentioned above.
    """
    name = "Geas"
    level = 5
    casting_time = "1 minute"
    casting_range = "60 feet"
    components = ('V',)
    materials = """"""
    duration = "30 days"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Cleric', 'Druid', 'Paladin', 'Wizard')


class GentleRepose(Spell):
    """You touch a corpse or other remains. For the duration, the target is protected 
    from decay and can’t become undead.
    
    The spell also effectively extends the time
     limit on raising the target from the dead, since days spent under the influence
     of this spell don’t count against the time limit of spells such as raise dead.
    """
    name = "Gentle Repose"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A pinch of salt and one copper piece placed on each of the corpse’s eyes, which must remain there for the duration"""
    duration = "10 days"
    ritual = True
    magic_school = "Necromancy"
    classes = ('Cleric', 'Wizard')


class GiantInsect(Spell):
    """You transform up to ten centipedes, three spiders, five wasps, or one scorpion 
    within range into giant versions of their natural forms for the duration. A 
    centipede becomes a giant centipede, a spider becaomes a giant spider, a wasp 
    becomes a giant wasp, and a scorpion becomes a giant scorpion.
    
    Each creature 
    obeys your verbal commands, and in combat, they act on your turn each round. The
     DM has the statistics for these creatures and resolves their actions and 
    movement.
    
    A creature remains in its giant size for the duration, until it drops
     to 0 hit points, or until you use an action to dismiss the effect on it.
    
    The 
    DM might allow you to choose different targets. For example, if you transform a 
    bee, its giant version might have the same statistics as a giant wasp.
    """
    name = "Giant Insect"
    level = 4
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid',)


class Glibness(Spell):
    """Until the spell ends, when you make a Charisma check, you can replace the number
     you roll with a 15. Additionally, no matter what you say, magic that would 
    determine if you are telling the truth indicates that you are being truthful.
    """
    name = "Glibness"
    level = 8
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Warlock')


class GlobeOfInvulnerability(Spell):
    """An immobile, faintly shimmering barrier springs into existence in a 10-foot 
    radius around you and remains for the duration.
    
    Any spell of 5th level or lower
     cast from outside the barrier can’t affect creatures or objects within it, even
     if the spell is cast using a higher level spell slot. Such a spell can target 
    creatures and objects within the barrier, but the spell has no effect on them. 
    Similarly, the area within the barrier is excluded from the areas affected by 
    such spells.
    
    At Higher Levels: When you cast this spell using a spell slot of 
    7th level or higher, the barrier blocks spells of one level higher for each slot
     level above 6th.
    """
    name = "Globe Of Invulnerability"
    level = 6
    casting_time = "1 action"
    casting_range = "Self (10-foot radius)"
    components = ('V', 'S', 'M')
    materials = """A glass or crystal bead that shatters when the spell ends"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Sorcerer', 'Wizard')


class GlyphOfWarding(Spell):
    """When you cast this spell, you inscribe a glyph that harms other creatures, 
    either upon a surface (such as a table or a section of floor or wall) or within 
    an object that can be closed (such as a book, a scroll, or a treasure chest) to 
    conceal the glyph.
    If you choose a surface, the glyph can cover an area of the 
    surface no larger than 10 feet in diameter. If you choose an object, that object
     must remain in its place; if the object is moved more than 10 feet from where 
    you cast this spell, the glyph is broken, and the spell ends without being 
    triggered.
    
    The glyph is nearly invisible and requires a successful Intelligence
     (Investigation) check against your spell save DC to be found.
    
    You decide what 
    triggers the glyph when you cast the spell. For glyphs inscribed on a surface, 
    the most typical triggers include touching or standing on the glyph, removing 
    another object covering the glyph, approaching within a certain distance of the 
    glyph, or manipulating the object on which the glyph is inscribed. For glyphs 
    inscribed within an object, the most common triggers include opening that 
    object, approaching within a certain distance of the object, or seeing or 
    reading the glyph. Once a glyph is triggered, this spell ends.
    
    You can further 
    refine the trigger so the spell activates only under certain circumstances or 
    according to physical characteristics (such as height or weight), creature kind 
    (for example, the ward could be set to affect aberrations or drow), or 
    alignment. You can also set conditions for creatures that don’t trigger the 
    glyph, such as those who say a certain password.
    
    When you inscribe the glyph, 
    choose explosive runes or a spell glyph.
    
    Explosive Runes
    When triggered, the 
    glyph erupts with magical energy in a 20-foot-radius sphere centered on the 
    glyph. The sphere spreads around corners. Each creature in the area must make a 
    Dexterity saving throw. A creature takes 5d8 acid, cold, fire, lightning, or 
    thunder damage on a failed saving throw (your choice when you create the glyph),
     or half as much damage on a successful one.
    
    Spell Glyph
    You can store a 
    prepared spell of 3rd level or lower in the glyph by casting it as part of 
    creating the glyph. The spell must target a single creature or an area. The 
    spell being stored has no immediate effect when cast in this way. When the glyph
     is triggered, the stored spell is cast. If the spell has a target, it targets 
    the creature that triggered the glyph. If the spell affects an area, the area is
     centered on that creature. If the spell summons hostile creatures or creates 
    harmful objects or traps, they appear as close as possible to the intruder and 
    attack it. If the spell requires concentration, it lasts until the end of its 
    full duration.
    
    At Higher Levels: 
    """
    name = "Glyph Of Warding"
    level = 3
    casting_time = "1 hour"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Incense and powdered diamond worth at least 200 gp, which the spell consumes"""
    duration = "Until dispelled or triggered"
    ritual = False
    magic_school = "Abjuration"
    classes = ('you', 'cast', 'this', 'spell', 'using', 'a', 'spell', 'slot', 'of', '4th', 'level', 'or', 'higher', 'the', 'damage', 'of', 'an', 'explosive', 'runes', 'glyph', 'increases', 'by', '1d8', 'for', 'each', 'slot', 'level', 'above', '3rd.', 'If', 'you', 'create', 'a', 'spell', 'glyph', 'you', 'can', 'store', 'any', 'spell', 'of', 'up', 'to', 'the', 'same', 'level', 'as', 'the', 'slot', 'you', 'use', 'for', 'the', 'glyph', 'of')


class Goodberry(Spell):
    """Up to ten berries appear in your hand and are infused with magic for the 
    duration. A creature can use its action to eat one berry. Eating a berry 
    restores 1 hit point, and the berry provides enough nourishment to sustain a 
    creature for one day.
    The berries lose their potency if they have not been 
    consumed within 24 hours of the casting of this spell.
    """
    name = "Goodberry"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A sprig of mistletoe"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Ranger')


class GraspingVine(Spell):
    """You conjure a vine that sprouts from the ground in an unoccupied space of your 
    choice that you can see within range. When you cast this spell, you can direct 
    the vine to lash out at a creature within 30 feet of it that you can see. That 
    creature must succeed on a Dexterity saving throw or be pulled 20 feet directly 
    toward the vine.
    
    Until the spell ends, you can direct the vine to lash out at 
    the same creature or another one as a bonus action on each of your turns.
    """
    name = "Grasping Vine"
    level = 4
    casting_time = "1 bonus action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Druid', 'Ranger')


class Grease(Spell):
    """Slick grease covers the ground in a 10-foot square centered on a point within 
    range and turns it into difficult terrain for the duration.
    
    When the grease 
    appears, each creature standing in its area must succeed on a Dexterity saving 
    throw or fall prone. A creature that enters the area or ends its turn there must
     also succeed on a Dexterity saving throw or fall prone.
    """
    name = "Grease"
    level = 1
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A bit of pork rind or butter"""
    duration = "1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Wizard',)


class GreaterInvisibility(Spell):
    """You or a creature you touch becomes invisible until the spell ends. Anything the
     target is wearing or carrying is invisible as long as it is on the target’s 
    person.
    """
    name = "Greater Invisibility"
    level = 4
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class GreaterRestoration(Spell):
    """You imbue a creature you touch with positive energy to undo a debilitating 
    effect. You can reduce the target’s exhaustion level by one, or end one of the 
    following effects on the target: 
    * One effect that charmed or petrified the 
    target 
    * One curse, including the target’s attunement to a cursed magic item 
    *
     Any reduction to one of the target’s ability scores 
    * One effect reducing the 
    target’s hit point maximum
    """
    name = "Greater Restoration"
    level = 5
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Diamond dust worth at least 100 gp, which the spell consumes"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Bard', 'Cleric', 'Druid')


class GreenFlameBlade(Spell):
    """As part of the action used to cast this spell, you must make a melee attack with
     a weapon against one creature within the spell's range, otherwise the spell 
    fails. On a hit, the target suffers the attack's normal effects, and green fire 
    leaps from the target to a different creature of your choice that you can see 
    within 5 feet of it. The second creature takes fire damage equal to your 
    spellcasting ability modifier. This spell's damage increases when you reach 
    higher levels.
    
    At Higher Levels: At 5th level, the melee attack deals an extra 
    1d8 fire damage to the target, and the fire damage to the second creature 
    increases to 1d8 + your spellcasting ability modifier. Both damage rolls 
    increase by 1d8 at 11th level and 17th level.
    """
    name = "Green-Flame Blade"
    level = 0
    casting_time = "1 action"
    casting_range = "5 feet"
    components = ('V', 'M')
    materials = """A weapon"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class GuardianOfFaith(Spell):
    """A Large spectral guardian appears and hovers for the duration in an unoccupied 
    space of your choice that you can see within range. The guardian occupies that 
    space and is indistinct except for a gleaming sword and shield emblazoned with 
    the symbol of your deity.
    
    Any creature hostile to you that moves to a space 
    within 10 feet of the guardian for the firs time on a turn must succeed on a 
    Dexterity saving throw. The creature takes 20 radiant damage on a failed save, 
    or half as much damage on a successful one. The guardian vanishes when it has 
    dealt a total of 60 damage.
    """
    name = "Guardian Of Faith"
    level = 4
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V',)
    materials = """"""
    duration = "8 hours"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Cleric',)


class GuardianOfNature(Spell):
    """A nature spirit answers your call and transforms you into a powerful guardian. 
    The transformation lasts until the spell ends. You choose one of the following 
    forms to assume: Primal Beast or Great Tree.
    Primal Beast. Bestial fur covers 
    your body, your facial features become feral, and you gain the following 
    benefits:
    - Your walking speed increases by 10 feet.
    - You gain darkvision with 
    a range of 120 feet.
    - You make Strength—based attack rolls with advantage.
    - 
    Your melee weapon attacks deal an extra 1d6 force damage on a hit.
    Great Tree. 
    Your skin appears barky, leaves sprout from your hair, and you gain the 
    following benefits:
    . You gain 10 temporary hit points.
    - You make Constitution 
    saving throws with advantage.
    - You make Dexterity- and Wisdom-based attack 
    rolls with advantage.
    - While you are on the ground, the ground within 15 feet 
    of you is difficult terrain for your enemies.
    """
    name = "Guardian Of Nature"
    level = 4
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Ranger')


class GuardsAndWards(Spell):
    """You create a ward that protects up to 2,500 square feet of floor space (an area 
    50 feet square, or one hundred 5-foot squares or twenty-five 10-foot squares). 
    The warded area can be up to 20 feet tall, and shaped as you desire. You can 
    ward several stories of a stronghold by dividing the area among them, as long as
     you can walk into each contiguous area while you are casting the spell.
    
    When 
    you cast this spell, you can specify individuals that are unaffected by any or 
    all of the effects that you choose. You can also specify a password that, when 
    spoken aloud, makes the speaker immune to these effects.
    
    Guards and wards 
    creates the following effects within the warded area.
    
    Corridors
    Fog fills all 
    the warded corridors, making them heavily obscured. In addition, at each 
    intersection or branching passage offering a choice of direction, there is a 50 
    percent chance that a creature other than you will believe it is going in the 
    opposite direction from the one it chooses.
    
    Doors
    All doors in the warded area 
    are magically locked, as if sealed by an arcane lock spell. In addition, you can
     cover up to ten doors with an illusion (equivalent to the illusory object 
    function of the m inor illusion spell) to make them appear as plain sections of 
    wall.
    
    Stairs
    Webs fill all stairs in the warded area from top to bottom, as the
     web spell. These strands regrow in 10 minutes if they are burned or torn away 
    while guards and wards lasts.
    
    Other Spell Effect
    You can place your choice of 
    one of the following magical effects within the warded area of the stronghold.
    •
     Place dancing lights in four corridors. You can designate a simple program that
     the lights repeat as long as
    guards and wards lasts.
    • Place magic mouth in two
     locations.
    • Place stinking cloud in two locations. The vapors appear in the 
    places you designate; they return within 10 minutes if dispersed by wind while 
    guards and wards lasts.
    • Place a constant gust of wind in one corridor or room.
    
    • Place a suggestion in one location. You select an area of up to 5 feet 
    square, and any creature that enters
    or passes through the area receives the 
    suggestion mentally.
    
    The whole warded area radiates magic. A dispel magic cast 
    on a specific effect, if successful, removes only that effect.
    You can create a 
    permanently guarded and warded structure by casting this spell there every 
    day for one year.
    """
    name = "Guards And Wards"
    level = 6
    casting_time = "10 minutes"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Burning incense, a small measure of brimstone and oil, a knotted string, a small amount of umber hulk blood, and a small silver rod worth at least 10 gp"""
    duration = "24 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Bard', 'Wizard')


class Guidance(Spell):
    """You touch one willing creature. Once before the spell ends, the target can roll 
    a d4 and add the number rolled to one ability check of its choice. It can roll 
    the die before or after making the ability check. The spell then ends.
    """
    name = "Guidance"
    level = 0
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Divination"
    classes = ('Cleric', 'Druid')


class GuidingBolt(Spell):
    """A flash of light streaks toward a creature of your choice within range.
    Make a 
    ranged spell attack against the target. On a hit, the target takes 4d6 radiant 
    damage, and the next attack roll made against this target before the end of your
     next turn has advantage, thanks to the mystical dim light glittering on the 
    target until then.
    
    At Higher Levels: When you cast this spell using a spell 
    slot of 2nd level or higher, the damage increases by 1d6 for each slot level 
    above 1st.
    """
    name = "Guiding Bolt"
    level = 1
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "1 round"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric',)


class Gust(Spell):
    """You seize the air and compel it to create one of the following effects at a 
    point you can see within range:
    • One Medium or smaller creature that you choose
     must succeed on a Strength saving throw or be pushed up to 5 feet away from 
    you.
    • You create a small blast of air capable of moving one object that is 
    neither held nor carried and that weighs no more than 5 pounds. The object is 
    pushed up to 10 feet away from you. It isn’t pushed with enough force to cause 
    damage.
    • You create a harmless sensory affect using air, such as causing leaves
     to rustle, wind to slam shutters shut, or your clothing to ripple in a breeze.
    """
    name = "Gust"
    level = 0
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class GustOfWind(Spell):
    """A line of strong wind 60 feet long and 10 feet wide blasts from you in a 
    direction you choose for the spell’s duration. Each creature that starts its 
    turn in the line must succeed on a Strength saving throw or be pushed 15 feet 
    away from you in a direction following the line.
    
    Any creature in the line must 
    spend 2 feet of movement for every 1 foot it moves when moving closer to you.
    
    
    The gust disperses gas or vapor, and it extinguishes candles, torches, and 
    similar unprotected flames in the area. It causes protected flames, such as 
    those of lanterns, to dance wildly and has a 50 percent chance to extinguish 
    them.
    
    As a bonus action on each of your turns before the spell ends, you can 
    change the direction in which the line blasts from you.
    """
    name = "Gust Of Wind"
    level = 2
    casting_time = "1 action"
    casting_range = "Self (60-foot line)"
    components = ('V', 'S', 'M')
    materials = """A legume seed"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Sorcerer', 'Wizard')





from .spells import Spell


class Bane(Spell):
    """Up to three creatures of your choice that you can see within range must make 
    Charisma saving throws. Whenever a target that fails this saving throw makes an 
    attack roll or a saving throw before the spell ends, the target must roll a d4 
    and subtract the number rolled from the attack roll or saving throw.
    
    At Higher 
    Levels: When you cast this spell using a spelslot of 2nd level or higher, you 
    can target one aditional creature for each slot level above 1st.
    """
    name = "Bane"
    level = 1
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A drop of blood"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Cleric')


class BanishingSmite(Spell):
    """The next time you hit a creature with a weapon attack before this spell ends, 
    your weapon crackles with force, and the attack deals an extra 5d10 force damage
     to the target. Additionally, if this attack reduces the target to 50 hit points
     of fewer, you banish it. If the target is native to a different plane of 
    existence than the on you’re on, the target disappears, returning to its home 
    plane. If the target is native to the plane you’re on, the creature vanishes 
    into a harmless demiplane. While there, the target is incapacitated. It remains 
    there until the spell ends, at which point the target reappears in the space it 
    left or in the nearest unoccupied space if that space is occupied.
    """
    name = "Banishing Smite"
    level = 5
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Paladin',)


class Banishment(Spell):
    """You attempt to send one creature that you can see within range to another place 
    of existence. The target must succeed on a Charisma saving throw or be banished.
     
    
    If the target is native to the plane of existence you’re on, you banish the 
    target to a harmless demiplane. While there, the target is incapacitated. The 
    target remains there until the spell ends, at which point the target reappears 
    in the space it left or in the nearest unoccupied space if that space is 
    occupied. 
    
    If the target is native to a different plane of existence that the 
    one you’re on, the target is banished with a faint popping noise, returning to 
    its home plane. 
    If the spell ends before 1 minute has passed, the target 
    reappears in the space it left or in the nearest unoccupied space if that space 
    is occupied. Otherwise, the target doesn’t return.
    
    At Higher Levels: When you 
    cast this spell using a spell slot of 5th level or higher, you can target one 
    additional creature for each slot level above 4th.
    """
    name = "Banishment"
    level = 4
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """An item distasteful to the target"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric', 'Paladin', 'Sorcerer', 'Warlock', 'Wizard')


class Barkskin(Spell):
    """You touch a willing creature. Until the spellends, the target’s skin has a 
    rough, bark-like appearance, and the target’s AC can’t be less than 16, 
    regardless of what kind of armor it is wearing.
    """
    name = "Barkskin"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A handful of oak bark"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Ranger')


class BeaconOfHope(Spell):
    """This spell bestows hope and vitality. Choose any number of creatures within 
    range. For the duration, each target has advantage on Wisdom saving throws and 
    death saving throws, and regains the maximum number of hit points possible from 
    any healing.
    """
    name = "Beacon Of Hope"
    level = 3
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric',)


class BeastBond(Spell):
    """You establish a telepathic link with one beast you touch that is friendly to you
     or charmed by you. The spell fails if the beast’s Intelligence is 4 or higher. 
    Until the spell ends, the link is active while you and the beast are within line
     of sight of each other. Through the link, the beast can understand your 
    telepathic messages to it, and it can telepathically communicate simple emotions
     and concepts back to you. While the link is active, the beast gains advantage 
    on attack rolls against any creature within 5 feet of you that you can see.
    """
    name = "Beast Bond"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A bit of fur wrapped in a cloth"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Divination"
    classes = ('Druid', 'Ranger')


class BeastSense(Spell):
    """You touch a willing beast. For the duration of the spell, you can use your 
    action to see through the beast’s eyes and hear what it hears, and continue to 
    do so until you use your action to return to your normal senses.
    """
    name = "Beast Sense"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('S',)
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = True
    magic_school = "Divination"
    classes = ('Druid', 'Ranger')


class BestowCurse(Spell):
    """You touch a creature, and that creature must succeed on a Wisdom saving throw or
     become cursed for the duration of the spell. When you cast this spell, choose 
    the nature of the curse from the following options:* Choose one ability score. 
    While cursed, the target has disadvantage on ability checks and saving throws 
    made with that ability score.* While cursed, the target has disadvantage on 
    attack rolls against you.* While cursed, the target must make a Wisdom saving 
    throw at the start of each of its turns. If it fails, it wastes its action that 
    turn doing nothing.* While the target is cursed, your attacks and spells deal an
     extra 1d8 necrotic damage to the target. 
    
    A remove curse spell ends this 
    effect. At the DM’s option, you may choose an alternative curse effect, but it 
    should be no more powerful than those described above.
    The DM has final say on 
    such a curse’s effect.
    
    At Higher Levels: If you cast this spell using a spell 
    slot of 4th level or higher, the duration is concentration, up to 10 minutes. 
    
    If you use a spell slot of 5th level or higher, the duration is 8 hours. 
    If 
    you use a spell slot of 7th level or higher, the duration is 24 hours. 
    If you 
    use a 9th level spell slot, the spell lasts until it is dispelled. 
    Using a 
    spell slot of 5th level or higher grants a duration that doesn’t require 
    concentration.
    """
    name = "Bestow Curse"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Bard', 'Cleric', 'Wizard')


class BigbysHand(Spell):
    """You create a Large hand of shimmering, translucent force in an unoccupied space 
    that you can see within range. The hand lasts for the spell’s duration, and it 
    moves at your command, mimicking the movements of your own hand. 
    
    The hand is 
    an object that has AC 20 and hit points equal to your hit point maximum. If it 
    drops to 0 hit points, the spell ends. It has a Strength of 26 (+8) and a 
    Dexterity of 10 (+0). The hand doesn’t fill its space. 
    
    When you cast the spell
     and as a bonus action on your subsequent turns, you can move the hand up to 60 
    feet and then cause one of the following effects with it.
    
    Clenched Fist 
    The 
    hand strikes one creature or object within 5 feet of it. Make a melee spell 
    attack for the hand using your game statistics. On a hit, the target takes 4d8 
    force damage. 
    
    Forceful Hand
    The hand attempts to push a creature within 5 feet
     of it in a direction you choose. Make a check with the hand’s Strength 
    contested by the Strength (Athletics) check of the target. If the target is 
    Medium or smaller, you have advantage on the check. If you succeed, the hand 
    pushes the target up to 5 feet plus a number of feet equal to five times your 
    spellcasting ability modifier. The hand moves with the target to remain within 5
     feet of it. 
    
    Grasping Hand
    The hand attempts to grapple a Huge or smaller 
    creature within 5 feet of it. You use the hand’s Strength score to resolve the 
    grapple. If the target is Medium or smaller, you have advantage on the check. 
    While the hand is grappling the target, you can use a bonus action to have the 
    hand crush it. When you do so, the target takes bludgeoning damage equal to 2d6 
    + your spellcasting ability modifier. 
    
    Interposing Hand
    The hand interposes 
    itself between you and a creature you choose until you give the hand a different
     command. The hand moves to stay between you and the target, providing you with 
    half cover against the target. The target can’t move through the hand’s space if
     its Strength score is less than or equal to the hand’s Strength score. If its 
    Strength score is higher than the hand’s Strength score, the target can move 
    toward you through the hand’s space, but that space is difficult terrain for the
     target.
    
    At Higher Levels: When you cast this spell using a spell slot of 6th 
    level or higher, the damage from the clenched fist option increases by 2d8 and 
    the damage from the grasping hand increases by 2d6 for each slot level above 
    5th.
    """
    name = "Bigbys Hand"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """An eggshell and a snakeskin glove"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Wizard',)


class BladeBarrier(Spell):
    """You create a vertical wall of whirling, razor-sharp blades made of magical 
    energy. The wall appears within range and lasts for the duration. You can make a
     straight wall up to 100 feet long, 20 feet high, and 5 feet thick, or a ringed 
    wall up to 60 feet in diameter, 20 feet high, and 5 feet thick. The wall 
    provides three-quarters cover to creatures behind it, and its space is difficult
     terrain. 
    
     When a creature enters the wall’s area for the first time on a turn
     or starts its turn there, the creature must make a Dexterity saving throw. On a
     failed save, the creature takes 6 d10 slashing damage. On a successful save, 
    the creature takes half as much damage.
    """
    name = "Blade Barrier"
    level = 6
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric',)


class BladeWard(Spell):
    """You extend your hand and trace a sigil of warding in the air. Until the end of 
    your next turn, you have resistance against bludgeoning, piercing, and slashing 
    damage dealt by weapon attacks.
    """
    name = "Blade Ward"
    level = 0
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "1 round"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class Bless(Spell):
    """You bless up to three creatures of your choice within range. Whenever a target 
    makes an attack roll or a saving throw before the spell ends, the target can 
    roll a d4 and add the number rolled to the attack roll or saving throw.
    
    At 
    Higher Levels: When you cast this spell using a spell slot of 2nd level or 
    higher, you can target one additional creature for each slot level above 1st.
    """
    name = "Bless"
    level = 1
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A sprinkling of holy water"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Cleric', 'Paladin')


class Blight(Spell):
    """Necromantic energy washes over a creature of your choice that you can see within
     range, draining moisture and vitality from it. The target must make a 
    Constitution saving throw. The target takes 8d8 necrotic damage on a failed 
    save, or half as much damage on a successful one. This spell has no effect on 
    undead or constructs. 
    
    If you target a plant creature or a magical plant, it 
    makes the saving throw with disadvantage, and the spell deals maximum damage to 
    it. 
    
    If you target a nonmagical plant that isn’t a creature, such as a tree or 
    shrub, it doesn’t make a saving throw; it simply withers and dies.
    
    At Higher 
    Levels: When you cast this spell using a spell slot of 5th level or higher, the 
    damage increases by 1d8 for each slot level above 4th.
    """
    name = "Blight"
    level = 4
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Druid', 'Sorcerer', 'Warlock', 'Wizard')


class BlindingSmite(Spell):
    """The next time you hit a creature with a melee weapon attack during this spell’s 
    duration, you weapon flares with a bright light, and the attack deals an extra 
    3d8 radiant damage to the target. Additionally, the target must succeed on a 
    Constitution saving throw or be blinded until the spell ends. 
    
    A creature 
    blinded by this spell makes another Constitution saving throw at the end of each
     of its turns. On a successful save, it is no longer blinded.
    """
    name = "Blinding Smite"
    level = 3
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Paladin',)


class Blindnessdeafness(Spell):
    """You can blind or deafen a foe. Choose one creature that you can see within range
     to make a Constitution saving throw. If it fails, the target is either blinded 
    or deafened (your choice) for the duration. At the end of each of its turns, the
     target can make a Constitution saving throw. On a success, the spell ends.
    
    At 
    Higher Levels: When you cast this spell using a spell slot of 3rd level or 
    higher, you can target one additional creature for each slot level above 2nd.
    """
    name = "Blindnessdeafness"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V',)
    materials = """"""
    duration = "1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Bard', 'Cleric', 'Sorcerer', 'Wizard')


class Blink(Spell):
    """Roll a d20 at the end of each of your turns for the duration of the spell. On a 
    roll of 11 or higher, you vanish from your current plane of existence and appear
     in the Etheral Plane (the spell fails and the casting is wasted if you were 
    already on that plane). 
    
    At the start of you next turn, and when the spell ends
     if you are on the Etheral Plane, you return to an unoccupied space of your 
    choice that you can see within 10 feet of the space you vanished from. If no 
    unoccupied space is available within that rang, you appear in the nearest 
    unoccupied space (chosen at random if more that one space is equally near). You 
    can dismiss this spell as an action. 
    
    While on the Ethereal Plane, you can see 
    and hear the plane you originated from, which is cast in shades of gray, and you
     can’t see anything more than 60 feet away.You can only affect and be affected 
    by other reatures on the Ethereal Plane. Creature that aren’t there can’t 
    perceive you or interact with you, unless they have the ability to do so.
    """
    name = "Blink"
    level = 3
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Wizard')


class Blur(Spell):
    """Your body becomes blurred, shifting and wavering to all who can see you. For the
     duration, any creature has disadvantage on attack rolls against you. An 
    attacker is immune to this effect if it doesnt rely on sight, as with 
    blindsight, or can see through illusions, as with truesight.
    """
    name = "Blur"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Sorcerer', 'Wizard')


class BonesOfTheEarth(Spell):
    """You cause up to six pillars of stone to burst from places on the ground that you
     can see within range. Each pillar is a cylinder that has a diameter of 5 feet 
    and a height of up to 30 feet. The ground where a pillar appears must be wide 
    enough for its diameter, and you can target ground under a creature if that 
    creature is Medium or smaller. Each pillar has AC 5 and 30 hit points. When 
    reduced to 0 hit points, a pillar crumbles into rubble, which creates an area of
     difficult terrain with a 10-foot radius. The rubble lasts until cleared.
    If a 
    pillar is created under a creature, that creature must succeed on a Dexterity 
    saving throw or be lifted by the pillar. A creature can choose to fail the save.
    
    If a pillar is prevented from reaching its full height because of a ceiling or 
    other obstacle, a creature on the pillar takes 6d6 bludgeoning damage and is 
    restrained, pinched between the pillar and the obstacle. The restrained creature
     can use an action to make a Strength or Dexterity check (the creature’s choice)
     against the spell’s saving throw DC. On a success, the creature is no longer 
    restrained and must either move off the pillar or fall off it.
    At Higher Levels.
     When you cast this spell using a spell slot of 7th level or higher, you can 
    create two additional pillars for each slot level above 6th.
    """
    name = "Bones Of The Earth"
    level = 6
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid',)


class BoomingBlade(Spell):
    """As part of the action used to cast this spell, you must make a melee attack with
     a weapon against one creature within the spell's range, otherwise the spell 
    fails.
    On a hit, the target suffers the attack's normal effects, and it becomes 
    sheathed in booming energy until the start of your next turn. If the target 
    willingly moves be- fore then, it immediately takes 1d8 thunder damage, and the 
    spell ends.
    This spell's damage increases when you reach higher levels.
    
    At 
    Higher Levels: At 5th level, the melee attack deals an extra 1d8 thunder damage 
    to the target, and the damage the target takes for moving increases to 2d8. Both
     damage rolls increase by 1d8 at 11th level and 17th level.
    """
    name = "Booming Blade"
    level = 0
    casting_time = "1 action"
    casting_range = "5 feet"
    components = ('V', 'M')
    materials = """A weapon"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class BrandingSmite(Spell):
    """The next time you hit a creature with a weapon attack before this spell ends, 
    the weapon glemas with astral radiance as you strike. The attack deals an extra 
    2d6 radiant damage to the target, which becomes visible if it is invisible, and 
    the target sheds dim light in a 5-foot radius and can’t become invisible until 
    the spell ends.
    
    At Higher Levels: When you cast this spell using a spell slot 
    of 3rd level or higher, the extra damage increases by 1d6 for each slot level 
    above 2nd.
    """
    name = "Branding Smite"
    level = 2
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Paladin',)


class BurningHands(Spell):
    """As you hold your hands with thumbs touching and fingers spread, a thin sheet of 
    flames shoots forth from your outstretched fingertips. Each creature in a 
    15-foot cone must make a Dexterity saving throw. A creature takes 3d6 fire 
    damage on a failed save, or half as much damage on a successful one. 
    
    The fire 
    ignites any flammable objects in the area that aren’t being worn or carried.
    
    At
     Higher Levels: When you cast this spell using a spell slot of 2nd level or 
    higher, the damage increases by 1d6 for each slot level above 1st.
    """
    name = "Burning Hands"
    level = 1
    casting_time = "1 action"
    casting_range = "Self (15-foot cone)"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')





class Shield():
    """A shield that can be worn on one hand."""
    name = "Shield"
    cost = "10 gp"
    base_armor_class = 2
    
    def __str__(self):
        return self.name

    def __repr__(self):
        return "\"{:s}\"".format(self.name)

    @classmethod
    def improved_version(cls, bonus):
        bonus = int(bonus)
        
        class NewShield(cls):
            name = f'+{bonus} ' + cls.name
            base_armor_class = cls.base_armor_class + bonus
            
        return NewShield


class WoodenShield(Shield):
    name = 'Wooden shield'


class ShieldOfFaces(Shield):
    name = "Shield +1"
    base_armor_class = 3


class NoShield(Shield):
    """If a character is carrying no shield."""
    name = "No shield"
    cost = "0"
    base_armor_class = 0
    
    def __str__(self):
        return self.name


class Armor():
    """A piece of armor that can be worn.
    
    Attributes
    ----------
    name : str
      Human-readable name for this armor.
    cost : str
      Cost and currency for this armor.
    base_armor_class : int
      Armor class granted before modifiers.
    dexterity_mod_max : int
      How much dexterity can the user contribute. ``0`` for no
      dexterity modifier, ``None`` for unlimited dexterity modifier.
    strength_required : int
      Minimum strength needed to use this armor properly.
    stealth_disadvantage : bool
      If true, the armor causes disadvantage on stealth rolls.
    weight_class : str
      light, medium, or heavy
    weight : int
      In lbs.
    
    """
    name = "Unknown Armor"
    cost = "0 gp"
    base_armor_class = 10
    dexterity_mod_max = None
    strength_required = None
    stealth_disadvantage = False
    weight = 0  # In lbs
    
    def __str__(self):
        return self.name

    def __repr__(self):
        return "\"{:s}\"".format(self.name)

    @classmethod
    def improved_version(cls, bonus):
        bonus = int(bonus)
        
        class NewArmor(cls):
            name = f'+{bonus} ' + cls.name
            base_armor_class = cls.base_armor_class + bonus
            
        return NewArmor


class NoArmor(Armor):
    name = "No Armor"


class LightArmor(Armor):
    name = "Light Armor"


class MediumArmor(Armor):
    name = "Medium Armor"


class HeavyArmor(Armor):
    name = "Heavy Armor"


class PaddedArmor(LightArmor):
    name = "Padded Armor"
    cost = "5 gp"
    base_armor_class = 11
    weight = 8
    stealth_disadvantage = True


class LeatherArmor(LightArmor):
    name = "Leather Armor"
    cost = "10 gp"
    base_armor_class = 11
    weight = 10


class StuddedLeatherArmor(LightArmor):
    name = "Studded Leather Armor"
    cost = "45 gp"
    base_armor_class = 12
    weight = 13


class HideArmor(MediumArmor):
    name = "Hide Armor"
    cost = "10 gp"
    base_armor_class = 12
    dexterity_mod_max = 2
    weight = 12


class ChainShirt(MediumArmor):
    name = "Chain Shirt"
    cost = "50 gp"
    base_armor_class = 13
    dexterity_mod_max = 2
    weight = 20


class ScaleMail(MediumArmor):
    name = "Scale Mail"
    cost = "50 gp"
    base_armor_class = 14
    dexterity_mod_max = 2
    stealth_disadvantage = True
    weight = 45


class Breastplate(MediumArmor):
    name = "Breastplate"
    cost = "400 gp"
    base_armor_class = 14
    dexterity_mod_max = 2
    weight = 20


class HalfPlate(MediumArmor):
    name = "Half Plate"
    cost = "750 gp"
    base_armor_class = 15
    dexterity_mod_max = 2
    stealth_disadvantage = True
    weight = 40


class RingMail(HeavyArmor):
    name = "Ring Mail"
    cost = "30 gp"
    base_armor_class = 14
    dexterity_mod_max = 0
    stealth_disadvantage = True
    weight = 40


class ChainMail(HeavyArmor):
    name = "Chain Mail"
    cost = "75 gp"
    base_armor_class = 16
    dexterity_mod_max = 0
    strength_required = 13
    stealth_disadvantage = True
    weight = 55


class SplintArmor(HeavyArmor):
    name = "Splint Armor"
    cost = "200 gp"
    base_armor_class = 17
    dexterity_mod_max = 0
    strength_required = 15
    stealth_disadvantage = True
    weight = 60


class PlateMail(HeavyArmor):
    name = "Plate Mail"
    cost = "1,500 gp"
    base_armor_class = 18
    dexterity_mod_max = 0
    strength_required = 15
    stealth_disadvantage = True
    weight = 65


# Custom Armor
class ElvenChain(MediumArmor):
    name = 'Elven Chain'
    cost = '5,000 gp'
    base_armor_class = 14
    dexerity_mod_max = 2
    weight = 20
    

light_armors = [PaddedArmor, LeatherArmor, StuddedLeatherArmor]
medium_armors = [HideArmor, ChainShirt, ScaleMail, Breastplate, HalfPlate]
heavy_armors = [RingMail, ChainMail, SplintArmor, PlateMail]
all_armors = light_armors + medium_armors + heavy_armors



from .features import Feature, FeatureSelector
from .ranger import Defense, Dueling
from .fighter import GreatWeaponFighting, Protection
from .. import (weapons, armor)


# PHB
class DivineSense(Feature):
    """The presence of strong evil registers on your senses like a noxious odor,
    and powerful good rings like heavenly music in your ears. As an action, you
    can open your awareness to detect such forces. Until the end of your next
    turn, you know the location of any celestial, fiend, or undead within 60
    feet of you that is not behind total cover. You know the type (celestial,
    fiend, or undead) of any being whose presence you sense, but not its
    identity (the vampire Count Strahd von Zarovich, for instance). Within the
    same radius, you also detect the presence of any place or object that has
    been consecrated or desecrated, as with the hallow spell.

    You can use this feature a number of times equal to 1 + your Charisma
    modifier. When you finish a long rest, you regain all expended uses.

    """
    _name = "Divine Sense"
    source = "Paladin"

    @property
    def name(self):
        num_uses = max(1, 1+self.owner.charisma.modifier)
        return self._name + ' ({:d}x/LR)'.format(num_uses)


class LayOnHands(Feature):
    """Your blessed touch can heal wounds. You have a pool of healing power that
    replenishes when you take a long rest. With that pool, you can restore a
    total number of hit points equal to your paladin level x 5.

    As an action, you can touch a creature and draw power from the pool to
    restore a number of hit points to that creature, up to the maximum amount
    remaining in your pool.

    Alternatively, you can expend 5 hit points from your pool of healing to
    cure the target of one disease or neutralize one poison affecting it. You
    can cure multiple diseases and neutralize multiple poisons with a single
    use of Lay on Hands, expending hit points separately for each one.

    This feature has no effect on undead and constructs

    """
    _name = "Lay on Hands"
    source = "Paladin"

    @property
    def name(self):
        level = self.owner.Paladin.level
        return self._name + " ({:d}HP/LR)".format(level*5)


class PaladinFightingStyle(FeatureSelector):
    """
    Select a Fighting Style by choosing in feature_choices:
 
    defense

    dueling

    great-weapon fighting

    protection
    """
    options = {'defense': Defense,
               'dueling': Dueling,
               'great': GreatWeaponFighting,
               'great-weapon fighting': GreatWeaponFighting,
               'projection': Protection}
    name = "Fighting Style (Select One)"
    source = "Paladin"


class DivineSmite(Feature):
    """Starting at 2nd level, when you hit a creature with a melee weapon attack,
    you can expend one paladin spell slot to deal radiant damage to the target,
    in addition to the weapon’s damage. The extra damage is 2d8 for a 1st-level
    spell slot, plus 1d8 for each spell level higher than 1st, to a maximum of
    5d8. The damage increases by 1d8 if the target is an undead or a fiend.

    """
    name = "Divine Smite"
    source = "Paladin"


class DivineHealth(Feature):
    """By 3rd level, the divine magic flowing through you makes you immune to
    disease """
    name = "Divine Health"
    source = "Paladin"


class ExtraAttackPaladin(Feature):
    """Beginning at 5th level, you can attack twice, instead of once, whenever you
    take the Attack action on your turn

    """
    name = "Extra Attack (2x)"
    source = "Paladin"


class AuraOfProtection(Feature):
    """Starting at 6th level, whenever you or a friendly creature within 10 feet
    of you must make a saving throw, the creature gains a bonus to the saving
    throw equal to your Charisma modifier (with a minimum bonus of +1). You
    must be conscious to grant this bonus.

    At 18th level, the range of this aura increases to 30 feet.

    """
    name = "Aura of Protection"
    source = "Paladin"


class AuraOfCourage(Feature):
    """Starting at 10th level, you and friendly creatures within 10 feet of you
    can’t be frightened while you are conscious.

    At 18th level, the range of this aura increases to 30 feet

    """
    name = "Aura of Courage"
    source = "Paladin"


class ImprovedDivineSmite(Feature):
    """By 11th level, you are so suffused with righteous might that all your melee
    weapon strikes carry divine power with them. Whenever you hit a creature
    with a melee weapon, the creature takes an extra 1d8 radiant damage. If you
    also use your Divine Smite with an attack, you add this damage to the extra
    damage of your Divine Smite.

    """
    name = "Improved Divine Smite"
    source = "Paladin"


class CleansingTouch(Feature):
    """Beginning at 14th level, you can use your action to end one spell on
    yourself or on one willing creature that you touch. You can use this
    feature a number of times equal to your Charisma modifier (a minimum of
    once). You regain expended uses when you finish a long rest.

    """
    _name = "Cleansing Touch"
    source = "Paladin"
    
    @property
    def name(self):
        num_uses = max(1, 1+self.owner.charisma.modifier)
        return self._name + ' ({:d}x/LR)'.format(num_uses)


class ChannelDivinityPaladin(Feature):
    """Your oath allows you to channel divine energy to fuel magical effects. Each
    Channel Divinity option provided by your oath explains how to use it.

    When you use your Channel Divinity, you choose which option to use. You
    must then finish a short or long rest to use your Channel Divinity
    again.

    Some Channel Divinity effects require saving throws.  When you use such an
    effect from this class, the DC equals your paladin spell save DC.

    """
    name = "Channel Divinity (1x/SR)"
    source = "Paladin"


# Oath of Devotion
class SacredWeapon(Feature):
    """As an action, you can imbue one weapon that you are holding with positive
    energy, using your Channel Divinity. For 1 minute, you add your Charisma
    modifier to attack rolls made with that weapon (with a minimum bonus of
    +1). The weapon also emits bright light in a 20-foot radius and dim light
    20 feet beyond that. If the weapon is not already magical, it becomes
    magical for the duration.

    You can end this effect on your turn as part of any other action. If you
    are no longer holding or carrying this weapon, or if you fall unconscious,
    this effect ends.

    """
    name = "Channel Divinity: Sacred Weapon"
    source = "Paladin (Oath of Devotion)"


class TurnTheUnholy(Feature):
    """As an action, you present your holy symbol and speak a prayer censuring
    fiends and undead, using your Channel Divinity. Each fiend or undead that
    can see or hear you within 30 feet of you must make a W isdom saving
    throw. If the creature fails its saving throw, it is turned for 1 minute or
    until it takes damage.

    A turned creature must spend its turns trying to move as far away from you
    as it can, and it can’t willingly move to a space within 30 feet of you. It
    also can’t take reactions. For its action, it can use only the Dash action
    or try to escape from an effect that prevents it from moving. If there’s
    nowhere to move, the creature can use the Dodge action.

    """
    name = "Channel Divinity: Turn the Unholy"
    source = "Paladin (Oath of Devotion)"


class AuraOfDevotion(Feature):
    """Starting at 7th level, you and friendly creatures within 10 feet of you
    can’t be charmed while you are conscious. At 18th level, the range of this
    aura increases to 30 feet.

    """
    name = "Aura of Devotion"
    source = "Paladin (Oath of Devotion)"


class PurityOfSpirit(Feature):
    """Beginning at 15th level, you are always under the effects of a protection
    from evil and good spell.

    """
    name = "Purity of Spirit"
    source = "Paladin (Oath of Devotion)"


class HolyNimbus(Feature):
    """At 20th level, as an action, you can emanate an aura of sunlight. For 1
    minute, bright light shines from you in a 30-foot radius, and dim light
    shines 30 feet beyond that.

    Whenever an enemy creature starts its turn in the bright light, the
    creature takes 10 radiant damage.

    In addition, for the duration, you have advantage on saving throws against
    spells cast by fiends or undead. Once you use this feature, you can’t use
    it again until you finish a long rest.

    """
    name = "Holy Nimbus"
    source = "Paladin (Oath of Devotion)"


# Oath of the Ancients
# Oath of Redemption
class EmissaryOfPeace(Feature):
    """You can use your Channel Divinity to augment your presence with divine
    power. As a bonus action, you grant yourself a +5 bonus to Charisma
    (Persuasion) checks for the next 10 minutes.

    """
    name = "Channel Divinity: Emissary of Peace"
    source = "Paladin (Oath of Redemption)"


class RebukeTheViolent(Feature):
    """You can use your Channel Divinity to rebuke those who use
    violence. Immediately after an attacker within 30 feet ofyou deals damage
    with an attack against a creature other than you, you can use your reaction
    to force the attacker to make a Wisdom saving throw. On a failed save, the
    attacker takes radiant damage equal to the damage it just dealt. On a
    successful save, it takes half as much damage.

    """
    name = "Channel Divinity: Rebuke the Violent"
    source = "Paladin (Oath of Redemption)"


class AuraOfTheGuardian(Feature):
    """Starting at 7th level, you can shield others from harm at the cost of your
    own health. When a creature within 10 feet of you takes damage, you can use
    your reaction to magically take that damage, instead of that creature tak—
    ing it. This feature doesn't transfer any other effects that might
    accompany the damage, and this damage can't be reduced in any way. At 18th
    level, the range of this aura increases to 30 feet.

    """
    name = "Aura of the Guardian"
    source = "Paladin (Oath of Redemption)"


class ProtectiveSpirit(Feature):
    """Starting at 15th level, a holy presence mends your wounds in battle. You
    regain hit points equal to 1d6 + half your paladin level if you end your
    turn in combat with fewer than half of your hit points remaining and you
    aren't incapacitated.

    """
    name = "Protective Spirit"
    source = "Paladin (Oath of Redemption)"


class EmissaryOfRedemption(Feature):
    """At 20th level, you become an avatar of peace, which gives you two benefits:
    
    --You have resistance to all damage dealt by other crea— tures (their
    attacks, spells, and other effects).

    --Whenever a creature hits you with an attack, it takes radiant damage
    equal to half the damage you take from the attack. 

    If you attack a creature, cast a spell on it, or deal damage to it by any
    means but this feature, neither benefit works against that creature until
    you finish a long rest

    """
    name = "Emissary of Redemption"
    source = "Paladin (Oath of Redemption)"
